/*
 * render.c -- enfle rendering engine
 * (C)Copyright 1998, 99 by Hiroshi Takekawa
 * This file is part of Enfle.
 *
 * Last Modified: Sun Feb 20 12:59:18 2000.
 * $Id: render.c,v 1.22 2000/02/20 04:07:31 sian Exp $
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "enfle.h"
#include "browse.h"
#include "cmrender.h"
#include "quantize.h"

static XImage *
create_ximage(Info *info, Binfo *binfo)
{
#ifdef MITSHM
  if (binfo->if_mitshm)
    return XShmCreateImage(info->disp, info->visual, info->depth, ZPixmap, NULL,
			   &binfo->shminfo, info->xwidth, info->xheight);
  else
#endif
    return XCreateImage(info->disp, info->visual, info->depth, ZPixmap, 0, NULL,
			info->xwidth, info->xheight, 8, 0);
}

static int x_shm_attach_failed;
#ifdef MITSHM
int
x_error_handler(Display *disp, XErrorEvent *err_ev)
{
  char buf[256];

  XGetErrorText(disp, err_ev->error_code, buf, 256);
  fprintf(stderr, "X Error(%d): %s\n", err_ev->error_code, buf);

  if (err_ev->error_code == BadAccess) {
    fprintf(stderr, "Probably XShmAttach() failed. Resuming...\n");
    x_shm_attach_failed = 1;
    return 0;
  }

  exit(1);
}
#endif

static void *
allocate_ximage(Info *info, Binfo *binfo, XImage **ximage_return)
{
  XImage *ximage = *ximage_return;

#ifdef MITSHM
  if (binfo->if_mitshm) {
    if ((binfo->shminfo.shmid = shmget(IPC_PRIVATE,
				       info->xdata_size, (IPC_CREAT | 0600))) < 0) {
      fprintf(stderr, "allocate_ximage: shmget failed\n");
      return NULL;
    }
    if ((binfo->shminfo.shmaddr = shmat(binfo->shminfo.shmid, 0, 0)) == (void *)-1) {
      fprintf(stderr, "allocate_ximage: shmat failed\n");
      return NULL;
    }
    binfo->shminfo.readOnly = False;
    if (binfo->if_warning)
      fprintf(stderr, "allocate_ximage: XShmAttach()'ing...\n");
    x_shm_attach_failed = 0;
    XSetErrorHandler(x_error_handler);
    XShmAttach(info->disp, &binfo->shminfo);
    XSync(info->disp, False);
    if (x_shm_attach_failed) {
      binfo->if_mitshm = 0;
      if (binfo->if_warning)
	fprintf(stderr, "allocate_ximage: XShmAttach failed, reverting to normal method.\n");
      XDestroyImage(ximage);
      if (binfo->if_warning)
	fprintf(stderr, "allocate_ximage: XDestroyImage() Ok.\n");
      shmctl(binfo->shminfo.shmid, IPC_RMID, NULL);
      shmdt(binfo->shminfo.shmaddr);
      *ximage_return = create_ximage(info, binfo);
      if (binfo->if_warning)
	fprintf(stderr, "allocate_ximage: Created normal XImage.\n");
      return calloc(1, info->xdata_size);
    }
    XSetErrorHandler(None);
    if (binfo->if_warning)
      fprintf(stderr, "allocate_ximage: XShmAttach()'d\n");
    shmctl(binfo->shminfo.shmid, IPC_RMID, NULL);
    return binfo->shminfo.shmaddr;
  } else
#endif
    return calloc(1, info->xdata_size);
}

void
render_destroy_ximage(Info *info, Binfo *binfo, XImage *ximage)
{
#ifdef MITSHM
  if (binfo->if_mitshm)
    XShmDetach(info->disp, &binfo->shminfo);
#endif
  XDestroyImage(ximage);
}

void
render_free_ximage(Info *info, Binfo *binfo)
{
#ifdef MITSHM
  if (binfo->if_mitshm)
    shmdt(binfo->shminfo.shmaddr);
  else
#endif
    free(info->xdata);
}

XImage *
render_image(Info *info, Binfo *binfo, Image *p, int f)
{
  XImage *ximage;
  int width = p->width;
  int height = p->height;
  int depth = info->depth;
  int i, j;
  unsigned char *xdata;

#ifdef DEBUG
  fprintf(stderr, "render_image() called with f = %d\n", f);
#endif
  if (!f) {
    info->xwidth = width + p->left;
    info->xheight = height + p->top;
    if (info->swidth)
      info->xwidth = info->mswidth;
    if (info->sheight)
      info->xheight = info->msheight;

    if ((ximage = create_ximage(info, binfo)) == NULL) {
      fprintf(stderr, "create_ximage() failed\n");
      exit(1);
    }

    info->xdata_size = ximage->bytes_per_line * ximage->height;
    if (info->xdata != NULL)
      render_free_ximage(info, binfo);

    if ((info->xdata = allocate_ximage(info, binfo, &ximage)) == NULL) {
      fprintf(stderr, "No enough memory xdata(%dx%d, depth = %d) %d bytes\n",
	      width, height, info->depth, info->xdata_size);
      exit(1);
    }
  } else {
    if ((ximage = create_ximage(info, binfo)) == NULL) {
      fprintf(stderr, "create_ximage() failed\n");
      exit(1);
    }

#ifdef MITSHM
    if (binfo->if_mitshm && !XShmAttach(info->disp, &binfo->shminfo)) {
      fprintf(stderr, "allocate_ximage: XShmAttach failed\n");
      shmdt(binfo->shminfo.shmaddr);
      exit(1);
    }
#endif
  }

  ximage->data = xdata = info->xdata;

#ifdef DEBUG
  fprintf(stderr, "type = %d default depth = %d\n", p->type, depth);
  fprintf(stderr, "(XImage's)bits_per_pixel = %d byte_order: %s bitmap_bit_order: %s\n",
	  ximage->bits_per_pixel,
	  (ximage->byte_order == LSBFirst) ? "LSBFirst" : "MSBFirst",
	  (ximage->bitmap_bit_order == LSBFirst) ? "LSBFirst" : "MSBFirst");
#endif

  /* expansion */
  switch (p->type) {
  case _MONO:
    if (depth != 1 && !image_expand(p, _INDEX)) {
      fprintf(stderr, "No enough memory for expand %d bytes\n", width * height);
      exit(1);
    }
    break;
  case _GRAY:
    if (!image_expand(p, _INDEX)) {
      fprintf(stderr, "Internal error: Error occurred during image_expand()\n");
      exit(1);
    }
    break;
  case _RGB16:
#ifdef DEBUG
    fprintf(stderr, "_RGB16\n");
#endif
    if (depth != 16 && depth != 15 && !image_expand(p, _RGB24)) {
      fprintf(stderr, "No enough memory for expand %d bytes\n", width * height * 3);
      exit(1);
    }
    p->ncolors = 65536;
    break;
  case _RGB24:
#ifdef DEBUG
    fprintf(stderr, "_RGB24\n");
#endif
    break;
  case _INDEX:
#ifdef DEBUG
    fprintf(stderr, "_INDEX\n");
#endif
    break;
  default:
    break;
  }

  /* reduction */
  if (!info->if_quantize) {
    switch (p->type) {
    case _RGB24:
      if (depth == 16 && !image_reduce(p, _RGB16)) {
	fprintf(stderr, "render_image: FATAL: image_reduce: returns 0\n");
	exit(1);
      } else if (depth == 15 && !image_reduce(p, _RGB15)) {
	fprintf(stderr, "render_image: FATAL: image_reduce: returns 0\n");
	exit(1);
      }
      break;
    default:
    }
  }

  /* rendering */
  if (!info->if_quantize) {
    switch (p->type) {
    case _MONO:
    case _GRAY:
      /* cannot reach here */
      fprintf(stderr, "Internal error, cannot rendering\n");
      exit(1);
    case _RGB16:
      if (depth != 16 && depth != 15) {
	/* cannot reach here */
	fprintf(stderr, "Internal error, cannot rendering\n");
	exit(1);
      }
      if (ximage->bits_per_pixel == 16) {
#ifdef MITSHM
	if (binfo->if_mitshm)
	  for (i = 0; i < height; i++)
	    memcpy(xdata + ximage->bytes_per_line * i, p->image + width * 2 * i, width * 2);
	else {
#endif
	  free(xdata);
	  info->xdata = NULL;
	  /* acceleration code */
	  ximage->data = xdata = p->image;
#ifdef MITSHM
	}
#endif
      } else {
	fprintf(stderr, "I have no code for bits_per_pixel = %d\n", ximage->bits_per_pixel);
	exit(1);
      }
      break;
    case _RGB24:
      if (depth != 24) {
	fprintf(stderr, "I have no code to render _RGB24 for depth %d\n", depth);
	exit(1);
      }
      if (ximage->bits_per_pixel == 24) {
	free(xdata);
	info->xdata = NULL;
#ifdef MITSHM
	if (binfo->if_mitshm)
	  for (i = 0; i < height; i++)
	    memcpy(xdata + ximage->bytes_per_line * i, p->image + width * 3 * i, width * 3);
	else
#endif
	  /* acceleration code */
	  ximage->data = xdata = p->image;
      } else if (ximage->bits_per_pixel == 32) {
	unsigned char *s = p->image;
	unsigned char *d = xdata;

	for (i = 0; i < width * height; i++) {
	  if (ximage->byte_order == LSBFirst) {
	    d[3] = 0;
	    d[2] = *s++;
	    d[1] = *s++;
	    d[0] = *s++;
	  } else {
	    d[0] = 0;
	    d[1] = *s++;
	    d[2] = *s++;
	    d[3] = *s++;
	  }
	  d += 4;
	}
      } else {
	fprintf(stderr, "I have no code for bits_per_pixel = %d\n", ximage->bits_per_pixel);
	exit(1);
      }
      break;
    default:
      break;
    }
  } else {
    if (p->ncolors > 256) {
      /* image quantization */
      unsigned char *q, pal[256 * 3];

      if ((q = quantize(p->image, pal, width, height, 8)) == NULL)
	return NULL;

      for (i = 0; i < 256; i++)
	for (j = 0; j < 3; j++)
	  p->colormap[i][j] = pal[i * 3 + j];
      free(p->image);
      p->image = q;
      p->type = _INDEX;
      p->bytes_per_line = p->width;
      p->ncolors = 256;
    }
  }

  if (p->type == _INDEX) {
    int f;

    /* colormapped image rendering */
    f = render_image_colormap(info, p, xdata, ximage->bits_per_pixel, ximage->bytes_per_line);
    if (f) {
      fprintf(stderr, "enfle: render_image: FATAL: %s\n", cmrender_err[f]);
      exit(1);
    }
  }

#ifdef SHAPE
  if (p->transparent_disposal == _SHAPE) {
    XImage *xmask;
    GC gc;
    XGCValues gcv;

    if (p->mask == NULL)
      image_create_mask(p);

    xmask = XCreateImage(info->disp, info->visual, 1, XYBitmap, 0, p->mask,
			 p->width, p->height, 8, 0);
    xmask->byte_order = MSBFirst;
    xmask->bitmap_bit_order = MSBFirst;
    p->pixmask = XCreatePixmap(info->disp, info->win, p->left + p->width, p->top + p->height, 1);
    gcv.background = 1;
    gcv.foreground = 0;
    gc = XCreateGC(info->disp, p->pixmask, GCBackground | GCForeground, &gcv);
    XFillRectangle(info->disp, p->pixmask, gc, 0, 0, p->left + p->width, p->top + p->height);
    XPutImage(info->disp, p->pixmask, gc, xmask, 0, 0, p->left, p->top, p->width, p->height);
    XFreeGC(info->disp, gc);
    XShapeCombineMask(info->disp, info->win, ShapeBounding, 0, 0, p->pixmask, ShapeSet);
    XDestroyImage(xmask);
    p->mask = NULL;
  }
#endif

  return ximage;
}
